import os, sys

project = u'Silo User Manual'
copyright = u'1996-2023, LLNL, LLNL-SM-654357'
author = u'LLNL'
release = u'4.11.1'

extensions = ['sphinx.ext.mathjax']

if os.environ.get('READTHEDOCS'):
    from subprocess import call
    call(['pip', 'install', 'myst_parser'])
extensions.append('myst_parser')

# Attempt add_css_file only if the version of Sphinx actually supports it
def setup(app):
    if hasattr(app, 'add_css_file'):
        app.add_css_file('custom.css')
    elif hasattr(app, 'add_stylesheet'):
        app.add_stylesheet('custom.css')

# Files to exclude
exclude_patterns = [ "func-template.md" ]

# Using default suffix would have been fine, but root_doc
# variable winds up assuming `.rst` without this.
source_suffix = {
    '.md': 'markdown'
}

# Top of MyST/Sphinx doc tree WITHOUT the extension
#root_doc = u'index-sphinx'

# MYST specific extensions
myst_enable_extensions = [
    "deflist",
    "dollarmath",
    "fieldlist",
    "substitution",
    "attrs_block",
    "attrs_inline",
    "colon_fence"
]
myst_heading_anchors = 3

myst_substitutions = {
  "EndFunc": "<hr class=\"docutils\" />"+"<br>"*40
}

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
try:
    html_theme
except:
    html_theme = "sphinx_rtd_theme"

html_static_path = ['_static']
html_theme_options = {
    # Toc options
    'navigation_depth': 4
}



#!/usr/bin/env python

import sys, os, re, math

hot_colors = {0:[0,0,255,255], 0.25:[0,255,255,255], 0.5:[0,255,0,255], 0.75:[255,255,0,255], 1:[255,0,0,255]}

# Va-Vb
def vecsub(Va,Vb):
    return [Va[0]-Vb[0],Va[1]-Vb[1],Va[2]-Vb[2]]

def vecmag(V):
    return math.sqrt(V[0]*V[0]+V[1]*V[1]+V[2]*V[2])

def vecdot(Va,Vb):
    return Va[0]*Vb[0]+Va[1]*Vb[1]+Va[2]*Vb[2]

def normalize(V):
    Vmag = vecmag(V)
    return [V[0]/Vmag, V[1]/Vmag, V[2]/Vmag]

def cross_prod(Va,Vb):
    return [Va[1]*Vb[2]-Va[2]*Vb[1],-(Va[0]*Vb[2]-Va[2]*Vb[0]),Va[0]*Vb[1]-Va[1]*Vb[0]]

def ray_tri_intersect(Pe, r, P1, P2, P3):
    V21 = vecsub(P2,P1)
    V32 = vecsub(P3,P2)
    Tn = cross_prod(V21,V32)
    tnum = Tn[0]*(P1[0]-Pe[0]) + Tn[1]*(P1[1]-Pe[1]) + Tn[2]*(P1[2]-Pe[2])
    tden = Tn[0]*r[0] + Tn[1]*r[1] + Tn[2]*r[2]
    if (tden == 0):
        return None
    t = tnum / tden
    # candidate point
    Pc = [Pe[0]+r[0]*t,Pe[1]+r[1]*t,Pe[2]+r[2]*t]

    # Now, check in/out of triangle using barycentric method
    V0 = vecsub(P3,P1)
    V1 = vecsub(P2,P1)
    V2 = vecsub(Pc,P1)
    dot00 = vecdot(V0,V0)
    dot01 = vecdot(V0,V1)
    dot02 = vecdot(V0,V2)
    dot11 = vecdot(V1,V1)
    dot12 = vecdot(V1,V2)
    invDenom = 1.0 / (dot00 * dot11 - dot01 * dot01)
    u = (dot11 * dot02 - dot01 * dot12) * invDenom
    v = (dot00 * dot12 - dot01 * dot02) * invDenom

    if u >= 0 and v >= 0 and (u+v) < 1:
        return Pc, u, v
    else:
        return None, 0, 0

def interp_field(u, v, A, B, C):
    return (1-u-v)*A + v*B + u*C

def interp_color(fval, fmin, fmax):
    ckeys = sorted(hot_colors.keys())
    cnum = len(ckeys)
    crange = ckeys[cnum-1] - ckeys[0]
    val = (fval-fmin)/(fmax-fmin) * crange + ckeys[0]
    if val <= ckeys[0]:
        return hot_colors[ckeys[0]]
    if val > ckeys[cnum-1]:
        return hot_colors[ckeys[cnum-1]]
    for i in range(1,len(ckeys)):
        if ckeys[i-1] <= val and val <= ckeys[i]:
            cval = (val-ckeys[i-1])/(ckeys[i]-ckeys[i-1])
            c0 = hot_colors[ckeys[i-1]]
            c1 = hot_colors[ckeys[i]]
            cdiff = vecsub(c1,c0)
            cdiff = [int(cdiff[0]*cval+c0[0]),int(cdiff[1]*cval+c0[1]),int(cdiff[2]*cval+c0[2])]
            return cdiff

def camera_params(LF, LA, VUP, FOV, W, H):
    N = [LA[0]-LF[0],LA[1]-LF[1],LA[2]-LF[2]]
    N = normalize(N)

    U = cross_prod(N,VUP)
    U = normalize(U)

    V = cross_prod(U,N)
    V = normalize(V)

    fova = math.tan(FOV/2)
    Xinc = [2*fova*U[0]/W, 2*fova*U[1]/W, 2*fova*U[2]/W]
    Yinc = [2*fova*V[0]/W, 2*fova*V[1]/W, 2*fova*V[2]/W]

    return N, U, V, Xinc, Yinc

def send_rays(LF, LA, VUP, FOV, W, H, P0, P1, P2, VMIN, VMAX):
    (N, U, V, Xinc, Yinc) = camera_params(LF, LA, VUP, FOV, W, H)

    for x in range(W):
        for y in range(H):
            R = [N[0] + 0.5*(2*y+1-H)*Yinc[0] + 0.5*(2*x+1-W)*Xinc[0],\
                 N[1] + 0.5*(2*y+1-H)*Yinc[1] + 0.5*(2*x+1-W)*Xinc[1],\
                 N[2] + 0.5*(2*y+1-H)*Yinc[2] + 0.5*(2*x+1-W)*Xinc[2]]
            R = normalize(R)
            ipoint_closests = None
            for i in range(12):
                (ipoint, u, v) = ray_tri_intersect(LF,R,P0[i],P1[i],P2[i])
                if ipoint != None:
                    if ipoint_closest != None:
                        if ipoint[2] > ipoint_closest[2]:
                            ipoint_closest = ipoint
                            fval = interp_field(u,v,P0[i][3],P1[i][3],P2[i][3])
                            rgb = interp_color(fval, VMIN, VMAX)
            print "%d:%d\t%d:%d:%d|%f"%(x,y,rgb[0],rgb[1],rgb[2],ipoint_closest[2])

def parse_zone_info(data):
    nodes = []
    nodeids = data[0].split(':')
    for n in nodeids:
        nspec = n.split('=')
        nodes.append(nspec[1])
    coords = []
    # data[1] = coords at each node
    coordlist = data[1].split(':')
    for coordval in coordlist:
        ctuple = coordval.split('$')
        coord = []
        for comp in ctuple:
            compval = comp.split('=')
            coord.append(compval[1])
        coords.append(coord)
    # data[2] = variables (both nodal and zonal)
    varlist = data[2].split(':')
    zonevars = {}
    nodevars = {}
    for varspec in varlist:
        vartuple = varspec.split('$')
        v = []
        if len(vartuple) == 1: # zone-centered case
            varvec = vartuple[0].split(';')
            if len(varvec) > 1: # vector var
                varval = varvec[0].split('=')
                varname = re.search("([^0-9]*)([0-9]{1})",varval[0])
                varname = varname.group(1)
                for varcomp in varvec:
                    varval = varcomp.split('=')
                    v.append(float(varval[1]))
            else: # scalar var
                varval = vartuple[0].split('=')
                varname = re.search("([^0-9]*)([0-9]{1})",varval[0])
                if varname == None:
                    continue
                varname = varname.group(1)
                v.append(float(varval[1]))
            zonevars[varname] = v
        else: # node-centered case
            for t in vartuple:
                varvec = t.split(';')
                if len(varvec) > 1: # vector var
                    varval = varvec[0].split('=')
                    varname = re.search("([^0-9]*)([0-9]{1})",varval[0])
                    varname = varname.group(1)
                    for varcomp in varvec:
                        varval = varcomp.split('=')
                        v.append(float(varval[1]))
                else: # scalar var
                    varval = varvec[0].split('=')
                    varname = re.search("([^0-9]*)([0-9]{1})",varval[0])
                    varname = varname.group(1)
                    v.append(float(varval[1]))
            nodevars[varname] = v
    return nodes, coords, zonevars, nodevars

def render_tri(coords, vals, na, nb, nc):
    p0 = [float(coords[na][0]),float(coords[na][1]),float(coords[na][2]),float(vals[na])]
    p1 = [float(coords[nb][0]),float(coords[nb][1]),float(coords[nb][2]),float(vals[nb])]
    p2 = [float(coords[nc][0]),float(coords[nc][1]),float(coords[nc][2]),float(vals[nc])]
    send_rays([0,0,10],[0,0,0],[0,1,0],0.5,20,20,p0,p1,p2,-10,10)

def render_zone(coords, vals):
    if len(coords) == 8: # hex
        render_tri(coords, vals, 0, 1, 5)
        render_tri(coords, vals, 0, 5, 4)

        render_tri(coords, vals, 0, 2, 1)
        render_tri(coords, vals, 0, 3, 2)

        render_tri(coords, vals, 0, 4, 7)
        render_tri(coords, vals, 0, 7, 3)
        
        render_tri(coords, vals, 1, 2, 6)
        render_tri(coords, vals, 1, 6, 5)

        render_tri(coords, vals, 2, 3, 7)
        render_tri(coords, vals, 2, 7, 6)

        render_tri(coords, vals, 4, 5, 6)
        render_tri(coords, vals, 4, 6, 7)

# input comes from STDIN (standard input)
for line in sys.stdin:

    # remove leading and trailing whitespace
    line = line.strip()

    # split the line into words
    (key, value) = line.split()

    data = value.split('|') # 1rst level split (nodes|coords|vars) sections

    (nodes, coords, zonevars, nodevars) = parse_zone_info(data)

    mode = os.environ['PLOTMODE']
    var = os.environ['PLOTVAR']

    if mode == "threshold":
        min = float(os.environ['THRESHOLD_MIN'])
        max = float(os.environ['THRESHOLD_MAX'])
        if var in nodevars:
            nodevals = nodevars[var]
            for val in nodevals:
                if val >= min and val <= max:
                    render_zone(coords, nodevals)
                    continue
        elif var in zonevars:
            val = zonevars[var]
            if val >= min and val <= max:
                render_zone(coords, val)
                continue
    elif mode == "iso-contour":
        isovals = os.eniron['ISOVALS'].split(':') 



#!/usr/bin/env python3

import Silo

# read box-in-a-box data
try:
    f = Silo.Open("box_in_a_box.silo", Silo.DB_READ)
except:
    f = Silo.Open("bin/box_in_a_box.silo", Silo.DB_READ)

msName = f.GetToc().qmesh_names[0]
msInfo = f.GetVarInfo(msName,1)

znName = f.GetToc().mat_names[0]
znInfo = f.GetVarInfo(znName, 1)

qvName = f.GetToc().qvar_names[0]
qvInfo = f.GetVarInfo(qvName, 1)

f.Close()

g = Silo.Create("scale_quadvar.silo", "scale data", Silo.DB_HDF5, Silo.DB_CLOBBER)

# write mesh data
g.WriteObject(msName, msInfo)

# write zone data
g.WriteObject(znName, znInfo)

# create scale data
scaleArr = tuple(val * 2 for val in qvInfo['value0'])

# override old quadvar with new data
qvInfo['name'] = "scaled_data"
qvInfo['value0'] = scaleArr

# write the new quadvar to the output file
g.WriteObject(qvInfo['name'], qvInfo)
# close the file
g.Close()



import exodus, Silo

#
# Main inputs to the script (values are just for example data)
#
siloFile = "globe.silo"
matobjName = "mat1"
matsOfInterest = [2, 3]
fieldsOfInterest =["dx", "u"] # one nodal and one zonal
exoFile = "s2ex-out.e"

# Open the silo file
silodb = Silo.Open(siloFile)

# Get the specified Silo material object and all its data
mat = silodb.GetVarInfo(matobjName, 1)

# Get the mesh object associated with this material object.
# including all its raw data (e.g. 1 for second arg)
mesh = silodb.GetVarInfo(mat["meshid"], 1)

# Get the Silo zonelist for this mesh
zl = silodb.GetVarInfo(mesh["zonelist"], 1)

#
# Traverse *both* zonelist and material objects simulatenously
# searching for zones containing materials of interest. Silo's
# zonelist is broken into 'segments' of constant shape size/type.
# Different segments can involve the same shape size/type though.
#
zloff = 0
zoneidx = 0
zoneids = []
newzl = {"nshapes":0, "shapecnt":[], "shapesize":[], "nodelist":[]}
for i in range(0,zl["nshapes"]):
    shsz = zl["shapesize"][i]
    for j in range(0,zl["shapecnt"][i]):
        keepZone = 0
        if mat["matlist"][zoneidx] in matsOfInterest: # zone with one material
            keepZone = 1
        elif mat["matlist"][zoneidx] < 0: # zone with multiple materials
            mixidx = -zoneidx-1
            while mat["mix_next"][mixidx] != 0:
                if mat["mix_mat"][mixidx] in matsOfInterest:
                    keepZone = 1
                    break
                mixidx = mat["mix_next"][mixidx]
        if keepZone:
            zoneids.append(zoneidx)
            k = len(newzl["shapesize"])
            if k == 0 or newzl["shapesize"][k-1] != shsz:
                newzl["shapesize"].append(shsz)
                newzl["shapecnt"].append(1)
                newzl["nshapes"] += 1
            else:
                newzl["shapecnt"][k-1] += 1
            newzl["nodelist"].extend(zl["nodelist"][zloff:zloff+shsz])
        zloff += shsz
        zoneidx += 1

#
# The zoneids list contains the set of zones we want
# Get unique set of node ids the newzl comprises
#
#uniqnl = sorted(list(set(newzl["nodelist"])))
uniqnl = list(set(newzl["nodelist"]))

#
# Map the newzl's nodelist node ids through the new unique set of nodes
#
newzl["nodelist"] = [uniqnl.index(x) for x in newzl["nodelist"]]

#
# Get coords for these nodes from the mesh
#
coord0 = [mesh["coord0"][x] for x in uniqnl]
coord1 = [mesh["coord1"][x] for x in uniqnl]
coord2 = [mesh["coord2"][x] for x in uniqnl]

#
# Map remaining fields of interest capturing all in fdata dict
# Note: If mapping non-scalar fields (e.g. vector or tensors),
# then there will be more 'valueX' entries in 'var' for each
# of the components
#
fdata = {"nodals":{}, "zonals":{}}
for f in fieldsOfInterest:
    var = silodb.GetVarInfo(f, 1)
    if var["centering"] == Silo.DB_ZONECENT:
        fdata["zonals"][f] = [var["value0"][x] for x in zoneids]
    else:
        fdata["nodals"][f] = [var["value0"][x] for x in uniqnl]

#
# Close the silo file
#
silodb.Close()


#
# Util funcs for converting silo to exodus
#
def Silo2ExodusElemType(sdim, nnodes):
    if sdim == 2:
        if nnodes == 3:
            return "TRI"
        elif nnodes == 4:
            return "QUAD"
    elif sdim == 3:
        if nnodes == 4:
            return "TET"
        elif nnodes == 5:
            return "PYR"
        elif nnodes == 6:
            return "WED"
        elif nnodes == 8:
            return "HEX"
    return "UNK"

def ConvertSiloWedgeOrderingToExodus(connarr):
#    s2eWedgeOrder = [5, 2, 0, 3, 4, 1]
    s2eWedgeOrder = [2, 5, 1, 3, 4, 0]
    nzones = len(connarr) / 6
    for i in range(0, nzones):
        nodes = connarr[i*6:i*6+6]
        connarr[i*6:i*6+6] = [nodes[x] for x in s2eWedgeOrder]

# Open the exodus output file and set some sizing information
e = exodus.exodus(exoFile, mode='w', title="generated by Silo2Exodus", \
    array_type='ctype', numDims=mesh["ndims"], numNodes=len(coord0),
    numElems=len(zoneids), numBlocks=newzl["nshapes"], numNodeSets=0, numSideSets=0)

# specify number of elem vars
e.set_element_variable_number(len(fdata["zonals"]))

# specify number of node vars
e.set_node_variable_number(len(fdata["nodals"]))

# Output coordinates
e.put_coords(coord0, coord1, coord2)

# Loop over the zonelist, building up args for put_concat_elem_blk
blkids=[]
elemtypes=[]
elemcnts=[]
elemsizes=[]
attrcnts=[]
for i in range(0,newzl["nshapes"]):
    blkids.append(i+1)
    # Map Silo zone shape info to exodus elem type info
    elemtypes.append(Silo2ExodusElemType(mesh["ndims"], newzl["shapesize"][i]))
    elemcnts.append(newzl["shapecnt"][i])
    elemsizes.append(newzl["shapesize"][i])
    attrcnts.append(0)

# Put out elem block header info
e.put_concat_elem_blk(blkids,elemtypes,elemcnts,elemsizes,attrcnts,0)

# Put out connectivity for each element block
offset=0
for i in range(0,newzl["nshapes"]):
    cnt = newzl["shapecnt"][i]*newzl["shapesize"][i]
    exconn = [nid+1 for nid in newzl["nodelist"][offset:offset+cnt]]
    if Silo2ExodusElemType(mesh["ndims"], newzl["shapesize"][i]) == "WED":
        ConvertSiloWedgeOrderingToExodus(exconn)
    e.put_elem_connectivity(i+1, exconn)
    offset += cnt

# Put out nodal variables
idx=1
for v in fdata["nodals"]:
    e.put_node_variable_name(v,idx)
    e.put_node_variable_values(v, 1, fdata["nodals"][v])
    idx += 1

# Put out zonal (elem) variables (upon each elem block)
idx=1
for v in fdata["zonals"]:
    offset=0
    e.put_element_variable_name(v,idx)
    for i in range(0,newzl["nshapes"]):
        cnt = newzl["shapecnt"][i]
        vals = [zid for zid in fdata["zonals"][v][offset:offset+cnt]]
        e.put_element_variable_values(i+1, v, 1, vals)
        offset += cnt
    idx += 1

e.close()



import re, itertools

#
# Map section headings to markdown file names
#
fileNameFromSec = {
    "Error Handling and Global Library Behavior": "globals",
    "Files and File Structure": "files",
    "Meshes, Variables and Materials": "objects",
    "Multi-Block Objects and Parallel I/O": "parallel",
    "Part Assemblies, AMR, Slide Surfaces,": "subsets",
    "Calculational and Utility": "utility",
    "Optlists": "optlists",
    "User Defined (Generic) Data and Objects": "generic",
    "JSON Interface to Silo Objects": "json",
    "Previously Undocumented Use Conventions": "conventions",
    "Fortran Interface": "fortran",
    "Python Interface": "python",
    "Deprecated Functions": "deprecated",
    "Silo Library Header File": "header"
}

# 
# Get all documentation lines from text file
#
def GetFileLines(fname):
    with open(fname) as txtfile:
        return txtfile.readlines()

def IsMethodHeader(i, lines):
    if i < len(lines)-1 and re.search(r'^—', lines[i+1]):
        return True
    return False

def IsAPISectionHeader(line):
    if re.search(r'^[0-9]*\s*API Section', line):
        return True
    return False

def FormatText(s, kwnames=[]):

    s = re.sub(r'“',r'"',s) # left double curly/smart
    s = re.sub(r'”',r'"',s) # right double curly/smart
    s = re.sub(r'‘',r"'",s) # left single curly/smart
    s = re.sub(r'’',r"'",s) # right single curly/smart
    s = re.sub(r'—',r'-',s) # emdash

    # Backtick any keyword words
    for kw in kwnames:
        s = re.sub(r'(\s+)%s(\s+|[\.?!])'%kw,r'\1`%s`\2'%kw,s)

    # Backtick any DBxxx... words
    s = re.sub(r'(\s+)DB([a-zA-Z0-9_]{3,})(\(\))?(\s+|[\.?!])',r'\1`DB\2\3`\4',s)

    # Backtick any all caps words
    s = re.sub(r'(\s+)([A-Z0-9_]{3,})(\s+|[\.?!])',r'\1`\2`\3',s)

    # Undo all-caps backticking for some special cases
    s = re.sub(r'`HDF(4|5)`',r'HDF\1',s)
    s = re.sub(r'`PDB`',r'PDB',s)
    s = re.sub(r'`NOT`',r'**not**',s)

    return s

#
# Handle various issues with typesetting an output line
#
def OutputLineWithFormatting(mdfile_lines, line, kwnames=[], indent=""):

    # Break text ending at elipsis, dot, question mark, exclamation point followed by
    line = re.sub(r'\(e\.g\. ',r'@e@g@@',line)
    line = re.sub(r'([a-zA-Z0-9_)]\.\.\.|[a-zA-Z0-9_)]\.|[a-zA-Z0-9_)]\?|[a-zA-Z0-9_)]!)\s+',r'\1\n',line).strip()
    line = re.sub(r'@e@g@@',r'(e.g. ',line)
    sentences = line.split('\n')

    for s in sentences:

        s = FormatText(s, kwnames)

        mdfile_lines.append(indent + s.strip())
        mdfile_lines.append("\n")

    mdfile_lines.append("\n")

#
# Begin a new API section (new md file)
#
def StartNewSection(i, s, lines, mdfiles):
    mdfile_lines = []
    hs = re.search(r'^[0-9]*\s*API Section\s*(.*)',lines[i])
    if len(hs.groups()) > 0:
        hdr = "%s"%FormatText(hs.groups()[0]).strip()
    else:
        hdr = "Section %d"%s
    sec = "API Section"
    mdfiles[hdr] = {sec:[]}
    s += 1
    mdfiles[hdr][sec].append("## %s\n"%hdr)
    mdfiles[hdr][sec].append("\n")
    i += 1

    while i < len(lines) and not IsMethodHeader(i, lines):
        OutputLineWithFormatting(mdfiles[hdr][sec], lines[i])
        i += 1

    return i, s, hdr

def ProcessSynopsis(mdfile_lines, i, lines):
    i += 1
    mdfile_lines.append("* **C Signature:**\n\n") # FIXME REMOVE THIS
    mdfile_lines.append("  ```\n")
    mdfile_lines.append('  ' + lines[i].strip())
    mdfile_lines.append("\n")
    i += 1
    while i < len(lines) and not \
        (IsMethodHeader(i, lines) or \
         re.search(r'^Fortran Equivalent:$', lines[i]) or \
         re.search(r'^Returns:$', lines[i]) or \
         re.search(r'^Description:$', lines[i]) or \
         re.search(r'^Arguments:$', lines[i])):
        mdfile_lines.append('      ' + lines[i].strip())
        mdfile_lines.append("\n")
        i += 1
    mdfile_lines.append("  ```\n")
    if re.search(r'^Fortran Equivalent:$', lines[i]):
        mdfile_lines.append("\n")
        i += 1
        if re.search(r'^None$', lines[i]):
            mdfile_lines.append("* **Fortran Signature:**\n\n")
            mdfile_lines.append("  ```\n")
            mdfile_lines.append("  None\n")
            mdfile_lines.append("  ```\n")
            i += 1
        else:
            mdfile_lines.append("* **Fortran Signature:**\n\n")
            mdfile_lines.append("  ```\n")
            fort_args = []
            fort_args_done = False
            while i < len(lines) and not \
                (IsMethodHeader(i, lines) or \
                 re.search(r'^Returns:$', lines[i]) or \
                 re.search(r'^Arguments:$', lines[i])):
                fort_args += [w.strip(' \n') for w in lines[i].split(',') if len(w.strip(' \n'))>0]
                if not fort_args_done and re.search(r'[ \t,a-zA-Z0-9_]*\)$', lines[i]):
                    n = 4
                    ident = "  "
                    while len(fort_args):
                        mdfile_lines.append(ident + ', '.join(fort_args[:n]))
                        if fort_args[:n][-1][-1] != ')':
                            mdfile_lines.append(",")
                        mdfile_lines.append("\n")
                        fort_args = fort_args[n:]
                        n = 6
                        ident = "     "
                    fort_args = []
                    fort_args_done = True
                    mdfile_lines.append("  ```\n\n")
                elif fort_args_done:
                    mdfile_lines.append('  ' + FormatText(lines[i].strip(' \n')))
                    mdfile_lines.append("\n")
                i += 1
    mdfile_lines.append("\n")
    return i

def ProcessArgumentListBlock(mdfile_lines, i, lines):
    i += 1
    args = []
    while i < len(lines) and not \
        (IsMethodHeader(i, lines) or \
         IsAPISectionHeader(lines[i]) or \
         re.search(r'^Returns:$', lines[i]) or \
         re.search(r'^Description:$', lines[i])):
        if lines[i].strip() != '':
            args += [lines[i]]
        i += 1
    if len(args) < 2:
        mdfile_lines.append("* **Arguments:**\n\n")
        mdfile_lines.append("  `None`\n")
        return i, []
    mdfile_lines.append("* **Arguments:**\n\n")
    if len(args) % 2 != 0:
        print("***ARGS PROBLEM***\n")
        print(args)
    # use non-breaking spaces to enforce min table width for args column
    mdfile_lines.append("  Arg&nbsp;name | Description\n")
    mdfile_lines.append("  :---|:---\n")
    argnames = [args[2*j].strip() for j in range(int(len(args)/2))]
    j = 0
    while j < len(args):
        mdfile_lines.append("  `%s` | %s\n"%(args[j].strip(), FormatText(args[j+1].strip(),argnames) if j+1<len(args) else "ARGS PROBLEM"))
        j += 2
    mdfile_lines.append("\n\n")
    return i, argnames

#https://github.com/LLNL/Silo/blob/mcm86-27jan22-convert-docs-fm2md/src/silo/silo.h.in#L802
#https://github.com/LLNL/Silo/blob/main/src/silo/silo.h.in#L802

def ProcessReturnBlock(mdfile_lines, i, lines):
    i += 1
    retlines = []
    while i < len(lines) and not \
        (IsMethodHeader(i, lines) or \
         IsAPISectionHeader(lines[i]) or \
         re.search(r'^Description:$', lines[i])):
        if lines[i].strip() != '' and lines[i].strip().lower()[:4] != 'none':
            retlines += [lines[i]]
        i += 1
    mdfile_lines.append("* **Returned value:**\n\n")
    if len(retlines):
        j = 0
        while j < len(retlines):
            OutputLineWithFormatting(mdfile_lines,retlines[j],[],"  ")
            j += 1
    else:
        mdfile_lines.append("  void");
    mdfile_lines.append("\n\n")
    return i

def ReadLinesInTabBlock(target_ntabs, i, desclines):

    complete_line = ''
    while complete_line.count('\t') < target_ntabs and i < len(desclines):
        complete_line += desclines[i]
        i += 1
    if complete_line.count('\t') != target_ntabs:
        print("Bad Case")
    complete_line = complete_line.strip()
    complete_line = re.sub("\n","<br>", complete_line)
    return i, [complete_line.split('\t')]

def OutputGatheredColumnsAsTable(mdfile_lines, cols, argnames=[], tabtitle=""):

    cols = list(map(list, itertools.zip_longest(*cols, fillvalue=None)))
    mdfile_lines.append("  **%s**\n\n"%(tabtitle+":" if tabtitle else "&nbsp;"))
    mdfile_lines.append('  ' + '|'.join(cols[0]))
    mdfile_lines.append("\n")
    mdfile_lines.append('  ' + ":---|"*(len(cols[0])-1))
    mdfile_lines.append(":---\n")

    for c in cols[1:]:
        c = [x if x else '' for x in c]
        mdfile_lines.append('  ' + FormatText('|'.join(c),argnames)) # FIXME
        mdfile_lines.append("\n")

    mdfile_lines.append("\n\n")

def ProcessDescription(mdfile_lines, i, lines, argnames=[]):

    i += 1
    desclines = []
    while i < len(lines) and \
        not IsMethodHeader(i, lines) and \
        not IsAPISectionHeader(lines[i]):
        desclines += [lines[i]]
        i += 1

    if not desclines:
        return i

    mdfile_lines.append("* **Description:**\n\n")

    j = 0
    target_ntabs = 0
    tabcols = []
    first_table = True
    while j < len(desclines):

        if target_ntabs == 0:

            if desclines[j].count('\t') == 0:

                # Output a normal "line" (which may be multiple sentences)
                # in description
                OutputLineWithFormatting(mdfile_lines, desclines[j], argnames, "  ")
                j += 1

            else:

                # Begin capturing a new set of lines for current table
                target_ntabs = desclines[j].count('\t')
                tabcols = [desclines[j].strip().split('\t')]
                j += 1

        else:

            if desclines[j].count('\t') == 0:

                # We've come to the end of current tab block.
                # First, output the current table.
                tabtitle = ""
                if first_table and argnames[-1] == 'optlist':
                    tabtitle = "Optlist options"
                OutputGatheredColumnsAsTable(mdfile_lines, tabcols, argnames, tabtitle)
                first_table = False
                tabcols = []

                # Indicate we're done with current table
                target_ntabs = 0

                # Now, output the current line
                OutputLineWithFormatting(mdfile_lines, desclines[j], argnames, "  ")
                j += 1

            else:

                # gather another "line" of current table
                j, newcol = ReadLinesInTabBlock(target_ntabs, j, desclines)
                tabcols += newcol

    if tabcols:
       
        tabtitle = ""
        if first_table and argnames[-1] == 'optlist':
            tabtitle = "Optlist options"
        OutputGatheredColumnsAsTable(mdfile_lines, tabcols, argnames, tabtitle)

    mdfile_lines.append("\n")

    return i

def ProcessMethod(i, lines, sec):
    meth = lines[i][:-1]
    sec[meth] = []
    sec[meth].append("### `%s()`\n\n"%meth)
    sec[meth].append("* **Summary:** %s\n\n"%FormatText(lines[i+1][1:].strip()))
    i += 2
    while i < len(lines) and \
        not IsMethodHeader(i, lines) and \
        not IsAPISectionHeader(lines[i]):
        argnames = []
        handled_line = False
        if re.search(r'^Synopsis:$', lines[i]):
            i = ProcessSynopsis(sec[meth], i, lines)
            handled_line = True
        if re.search(r'^Arguments:$', lines[i]):
            i, argnames = ProcessArgumentListBlock(sec[meth], i, lines)
            handled_line = True
        if re.search(r'^Returns:$', lines[i]):
            i = ProcessReturnBlock(sec[meth], i, lines)
            handled_line = True
        if re.search(r'^Description:$', lines[i]):
            i = ProcessDescription(sec[meth], i, lines, argnames)
            handled_line = True
        if not handled_line:
            print("Didn't classify line...\"",lines[i],"\"")
            OutputLineWithFormatting(sec[meth], lines[i], argnames, "  ")
            i += 1
    sec[meth].append("---\n")
    sec[meth].append('<br>'*40)
    sec[meth].append("\n")
    return i, meth

#
# main program
#
lines = GetFileLines("Chapter2-man_pages2.txt")

i = 0
s = 0
mdfiles = {}
while i < len(lines):
    handled_line = False
    if IsAPISectionHeader(lines[i]):
        i, s, sec = StartNewSection(i, s, lines, mdfiles)
        handled_line = True
    if IsMethodHeader(i, lines):
        i, m = ProcessMethod(i, lines, mdfiles[sec])
        handled_line = True
    if not handled_line:
        print("Didn't classify line...\"",lines[i],"\"")
        OutputLineWithFormatting(lines[i], mdfiles[sec][m])
        i += 1

#
# Make a pass creating links from text of the form...
#    (See DBOpen)
#    See "DBSet...
#    See `DBReadVar` 
#    See “DBMakeMrgtree” on page 196
#    See below
#    See description
#    See documentation
#    See HDF5 reference manual
#
usingSec=""
def fixlinks(match):

    func = match.groups()[2]+match.groups()[3]
    lfunc = func.lower()

    for sec in mdfiles:
        if lfunc in [meth.lower() for meth in mdfiles[sec].keys()]:
            if sec == usingSec:
                return "See [`"+func+"`](#"+lfunc+")"
            else:
                return "See [`"+func+"`](./"+fileNameFromSec[sec]+".md#"+lfunc+")"

    return "See [`"+func+"`](UNKNOWN)"

for sec in mdfiles:
    for meth in mdfiles[sec]:
        for i in range(len(mdfiles[sec][meth])):

            #
            # skip lines that don't have 'See' indicating a ref
            #
            if 'See' not in mdfiles[sec][meth][i]:
                continue 

            #
            # Handle links to HDF5 docs
            #
            mdfiles[sec][meth][i] = re.sub(r'See HDF5 reference manual',
                r'See [HDF5 reference manual](https://docs.hdfgroup.org/hdf5/develop/)',mdfiles[sec][meth][i])

            #
            # Handle internal links
            #
            usingSec = sec
            mdfiles[sec][meth][i] = re.sub(r'See (the description of |the write-up of |the documentation for |the documentation on |documentation on )?(["`]?)(DB|db)([a-zA-Z0-9_]*)(["`]?)( on page ([0-9-]*))?',
                fixlinks, mdfiles[sec][meth][i])


def ProcessMethodTable(mdfile, methods):

    mdfile.write("### Methods and symbols in this section\n")
    mdfile.write("\n")
    mdfile.write("&nbsp;|&nbsp;|&nbsp;\n")
    mdfile.write(":---|:---|:---\n")

    i = 1
    for m in methods:
        if m == "API Section":
            continue
        mdfile.write("[`"+m+"`](#"+m.lower()+")")
        if i%3 == 0:
            mdfile.write("\n")
        else:
            mdfile.write("|")
        i += 1
    mdfile.write("\n")
    mdfile.write("\n")

    mdfile.write("---\n")
    mdfile.write('<br>'*40)
    mdfile.write("\n")
    
#
# Output the files
#
s = 0
for sec in mdfiles:
#    with open("Chapter2-Section%d.md"%s,"wt") as mdfile:
    with open("%s.md"%fileNameFromSec[sec],"wt") as mdfile:
        for meth in mdfiles[sec]:
            mdfile.writelines(mdfiles[sec][meth])
            if meth == "API Section":
                ProcessMethodTable(mdfile, sorted(mdfiles[sec].keys()))
    s += 1



#!/usr/bin/env python
 
from operator import itemgetter
import sys
 
current_word = None
current_z = None
current_rgb = None
word = None
 
# input comes from STDIN
for line in sys.stdin:
    # remove leading and trailing whitespace
    line = line.strip()
 
    # parse the input we got from mapper.py
    word, pixval = line.split('\t', 1)
 
    rgbz = pixval.split('|')
    rgb = rgbz[0]
    z = rgbz[1]

    # convert count (currently a string) to int
    try:
        z = float(z)
    except ValueError:
        # count was not a number, so silently
        # ignore/discard this line
        continue
 
    # this IF-switch only works because Hadoop sorts map output
    # by key (here: word) before it is passed to the reducer
    if current_word == word:
        if z < current_z:
            current_z = z
            curent_rgb = rgb
    else:
        if current_word:
            # write result to STDOUT
            print '%s\t%s' % (current_word, current_rgb)
        current_word = word
        current_rgb = rgb 
        current_z = z 
 
# do not forget to output the last word if needed!
if current_word == word:
    print '%s\t%s' % (current_word, current_rgb)



# Copyright (C) 1994-2016 Lawrence Livermore National Security, LLC.
# LLNL-CODE-425250.
# All rights reserved.
# 
# This file is part of Silo. For details, see silo.llnl.gov.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the disclaimer below.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the disclaimer (as noted
#      below) in the documentation and/or other materials provided with
#      the distribution.
#    * Neither the name of the LLNS/LLNL nor the names of its
#      contributors may be used to endorse or promote products derived
#      from this software without specific prior written permission.
# 
# THIS SOFTWARE  IS PROVIDED BY  THE COPYRIGHT HOLDERS  AND CONTRIBUTORS
# "AS  IS" AND  ANY EXPRESS  OR IMPLIED  WARRANTIES, INCLUDING,  BUT NOT
# LIMITED TO, THE IMPLIED  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A  PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN  NO  EVENT SHALL  LAWRENCE
# LIVERMORE  NATIONAL SECURITY, LLC,  THE U.S.  DEPARTMENT OF  ENERGY OR
# CONTRIBUTORS BE LIABLE FOR  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR  CONSEQUENTIAL DAMAGES  (INCLUDING, BUT NOT  LIMITED TO,
# PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS  OF USE,  DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER  IN CONTRACT, STRICT LIABILITY,  OR TORT (INCLUDING
# NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT  OF THE USE  OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# 
# This work was produced at Lawrence Livermore National Laboratory under
# Contract  No.   DE-AC52-07NA27344 with  the  DOE.  Neither the  United
# States Government  nor Lawrence  Livermore National Security,  LLC nor
# any of  their employees,  makes any warranty,  express or  implied, or
# assumes   any   liability   or   responsibility  for   the   accuracy,
# completeness, or usefulness of any information, apparatus, product, or
# process  disclosed, or  represents  that its  use  would not  infringe
# privately-owned   rights.  Any  reference   herein  to   any  specific
# commercial products,  process, or  services by trade  name, trademark,
# manufacturer or otherwise does not necessarily constitute or imply its
# endorsement,  recommendation,   or  favoring  by   the  United  States
# Government or Lawrence Livermore National Security, LLC. The views and
# opinions  of authors  expressed  herein do  not  necessarily state  or
# reflect those  of the United  States Government or  Lawrence Livermore
# National  Security, LLC,  and shall  not  be used  for advertising  or
# product endorsement purposes.
import Silo
import os

db = Silo.Create("foo.silo", "test file")
db.Write("t1", 12)
db.Write("t2", "hello")
db.Write("t3", (13,))
db.Write("t4", 14.5)
try:
    db.Write("t5", (1.2, 10))
except TypeError:
    pass
db.Write("t5", (1.2, 10.2))
db.Write("t6", (1.2, 10.2))
try:
    db.Write("t7", (10, 1.2))
except TypeError:
    pass
db.Write("t7", (10, 12))
db.Write("t8", "a")
db.MkDir("a")
db.Write("a/t8", "x2")
db.SetDir("a")
db.Write("t9", "x3")
db.Write("../t10", "x4")
db.SetDir("..")
db.Write("t11", "x5")
db.Write("t12",(1,2,3,4,5,6),(2,3),Silo.DB_INT)
db.Write("t13",(1.1,2.2,3.3,4.4,5.5,6.6),(3,2),Silo.DB_FLOAT)
db.Write("t14",(1.1,2.2,3.3,4.4,5.5,6.6,7.7,8.8,9.9),(3,3),Silo.DB_DOUBLE)
db.Close()

db2=Silo.Open("foo.silo")
print("toc=%s"%db2.GetToc())
print("t2='%s'"%db2.GetVar("t2"))
print("t6=%.2f,%.2f"%db2.GetVar("t6"))
print("a/t8='%s'"%db2.GetVar("a/t8"))
db2.SetDir("a")
print("t9=%s"%db2.GetVar("t9"))
print("/t5=%.2f,%.2f"%db2.GetVar("../t5"))
db2.Close()

#
# Now, test reading a raw DB_FLOAT array and overwritting it
#

# Lets get original size of file before the overwrite
origSize = os.path.getsize("sami.silo")

# Open file for overwrite
db3 = Silo.Open("sami.silo", Silo.DB_APPEND)

#
# This call gets the "raw" array as a dict holding
# 'dims', 'ndims', 'datatype' and 'data'
#
dv = db3.GetVarInfo("4d_float", 1)

# Ok, lets create alternative data
dvl = [x*3.1415926 if x < 65 else x for x in dv['data']]

# Now, overwrite the original data in the file
db3.Write("4d_float",dvl,dv['dims'],dv['datatype'])

db3.Close()

newSize = os.path.getsize("sami.silo")
assert newSize == origSize



# Copyright (C) 1994-2016 Lawrence Livermore National Security, LLC.
# LLNL-CODE-425250.
# All rights reserved.
# 
# This file is part of Silo. For details, see silo.llnl.gov.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the disclaimer below.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the disclaimer (as noted
#      below) in the documentation and/or other materials provided with
#      the distribution.
#    * Neither the name of the LLNS/LLNL nor the names of its
#      contributors may be used to endorse or promote products derived
#      from this software without specific prior written permission.
# 
# THIS SOFTWARE  IS PROVIDED BY  THE COPYRIGHT HOLDERS  AND CONTRIBUTORS
# "AS  IS" AND  ANY EXPRESS  OR IMPLIED  WARRANTIES, INCLUDING,  BUT NOT
# LIMITED TO, THE IMPLIED  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A  PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN  NO  EVENT SHALL  LAWRENCE
# LIVERMORE  NATIONAL SECURITY, LLC,  THE U.S.  DEPARTMENT OF  ENERGY OR
# CONTRIBUTORS BE LIABLE FOR  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR  CONSEQUENTIAL DAMAGES  (INCLUDING, BUT NOT  LIMITED TO,
# PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS  OF USE,  DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER  IN CONTRACT, STRICT LIABILITY,  OR TORT (INCLUDING
# NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT  OF THE USE  OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# 
# This work was produced at Lawrence Livermore National Laboratory under
# Contract  No.   DE-AC52-07NA27344 with  the  DOE.  Neither the  United
# States Government  nor Lawrence  Livermore National Security,  LLC nor
# any of  their employees,  makes any warranty,  express or  implied, or
# assumes   any   liability   or   responsibility  for   the   accuracy,
# completeness, or usefulness of any information, apparatus, product, or
# process  disclosed, or  represents  that its  use  would not  infringe
# privately-owned   rights.  Any  reference   herein  to   any  specific
# commercial products,  process, or  services by trade  name, trademark,
# manufacturer or otherwise does not necessarily constitute or imply its
# endorsement,  recommendation,   or  favoring  by   the  United  States
# Government or Lawrence Livermore National Security, LLC. The views and
# opinions  of authors  expressed  herein do  not  necessarily state  or
# reflect those  of the United  States Government or  Lawrence Livermore
# National  Security, LLC,  and shall  not  be used  for advertising  or
# product endorsement purposes.
import Silo

db = Silo.Create("onehexpy.silo", "This is a test to see how well this string displays in Qt widget in Silex\nThis is a test to see how well this string displays in Qt widget in Silex\nThis is a test to see how well this string displays in Qt widget in Silex\nThis is a test to see how well this string displays in Qt widget in Silex\n")

# coordinate arrays for the mesh
coord0 = (0, 0.707107, 0, -0.707107, 0, 0.707107, 0, -0.707107)
coord1 = (0, 0.5, 1, 0.5, 0.707107, 1.20711, 1.70711, 1.20711)
coord2 = (0, 0.5, 1, 0.5, -0.707107, -0.207107, 0.292893, -0.207107)
min_extents = (-0.707107, 0, -0.707107)
max_extents = (0.707107, 1.70711, 1)

# the ucd mesh object
mesh = {"name":"hex",\
        "type":"DBucdmesh",\
        "coord0":coord0,\
        "coord1":coord1,\
        "coord2":coord2,\
        "min_extents":min_extents,\
        "max_extents":max_extents,\
        "zonelist":"zonelist",\
        "ndims":3,\
        "nnodes":8,\
        "nzones":1,\
        "facetype":100,\
        "cycle":0,\
        "coord_sys":Silo.DB_OTHER,\
        "datatype":Silo.DB_DOUBLE,\
        "origin":0}
db.WriteObject("hex", mesh)

# nodelist and shape arrays for the zonelist
nodelist = (0, 1, 2, 3, 4, 5, 6, 7)
shapecnt = (1,)
shapesize = (8,)

# zonelist object
zonelist = {"name":"zonelist",\
            "type":"DBzonelist",\
            "ndims":3,\
            "nzones":1,\
            "nshapes":1,\
            "shapecnt":shapecnt,\
            "shapesize":shapesize,\
            "lnodelist":len(nodelist),\
            "nodelist":nodelist}
db.WriteObject("zonelist", zonelist)

# variable array
value0 = (0, 0.957107, 1, -0.457107, -0.5, 0.457107, 0.5, -0.957107)

# variable object
v = {"name":"v",\
     "type":"DBucdvar",\
     "meshid":"hex",\
     "datatype":Silo.DB_DOUBLE,\
     "nels":len(value0),\
     "nvals":1,\
     "ndims":3,\
     "centering":Silo.DB_NODECENT,\
     "value0":value0}
db.WriteObject("v", v)

db.Close()
